# -*- coding: utf-8 -*-

from __future__ import print_function

import argparse as ap
import json
import logging
import os

import pika
import sys
import time

from pyat import version
from pyat.pymygdala import LogHandler, RPCEvent
from pycapo import CapoConfig

_DESCRIPTION = """AAT/PPI workflow launcher, version {}. Launch a workflow from 
the command line."""


def _make_parser():
    r""" Build a command line parser for this app: this is external to MrClean and/or CapoSettings because both need
    it, and Sphinx will want a look at ot to build docs. """
    result = ap.ArgumentParser(description=_DESCRIPTION.format(version),
                               formatter_class=ap.RawTextHelpFormatter)
    result.add_argument('-P', '--profile', action='store',
                        help='profile name to use, e.g. test, production')
    result.add_argument('-e', '--exchange', action='store', default='archive.workflow-commands',
                        help='exchange to send to (you probably want the default)')
    result.add_argument('-s', '--processingsite', action='store', default='DSOC', choices=['NAASC', 'DSOC'],
                        help='site where processing should occur (either DSOC or NAASC)')
    result.add_argument('-p', '--property', action='append', default=[],
                        help="provide a KEY=VALUE pair to be passed to the workflow; can be used multiple times")
    result.add_argument('workflow', action='store', help='name of the workflow to initiate; e.g. runDownloadWorkflow, runSdmIngestionWorkflow')
    return result

def wf(**kwargs):
    # parse out the command line args here and decide what to do
    args = _make_parser().parse_args(**kwargs)

    connection = connect_to_amqp(args.profile)
    channel = connection.channel()

    # make the exchange we need
    channel.exchange_declare(args.exchange, 'topic', durable=True)

    #process the properties passed in
    properties = dict(arg.split('=') for arg in args.property)

    # if it's in there, fileSetIds needs some special care.
    if 'fileSetIds' in properties:
        #The script needs to pass an array in the json.  This drop spare square brackets
        #and split according to commas to create a list which is then decoded properly
        middleman = properties['fileSetIds']
        #replace any square braces in the input
        idList = middleman.replace('[','').replace(']','').split(',')
        #make fileSetIds a list of strings, not a string with a list
        properties['fileSetIds'] = idList

    # if it's in there, the bdf list needs some special care.
    if 'bdfs' in properties:
        #The script needs to pass an array in the json.  This drop spare square brackets
        #and split according to commas to create a list which is then decoded properly
        middleman = properties['bdfs']
        #replace any square braces in the input
        idList = middleman.replace('[','').replace(']','').split(',')
        #make fileSetIds a list of strings, not a string with a list
        properties['bdfs'] = idList

    # The sets of imaging parameters need to be a JsonArray, so we have work to do:
    if 'optimizedImagingParameters' in properties:
        cube_list = []
        json_parameters = json.loads(properties['optimizedImagingParameters'])

        # create a list of cubes, which will get interpreted correctly:
        for cube in json_parameters:
            cube_list.append(cube)

        # replace the old string we were passed with the new value
        properties['optimizedImagingParameters'] = cube_list

    # build the message
    message = {'eventName': args.workflow,
               'type': 'edu.nrao.archive.workflow.messaging.commands.StartWorkflow',
               'additionalPaths': [],
               'metadata': properties}

    # dispatch it
    channel.basic_publish(exchange=args.exchange,
                          routing_key=args.processingsite.lower() + '.start',
                          body=json.dumps(message))


def connect_to_amqp(profile):
    # load up Capo
    capo = CapoConfig(profile=profile)

    # get AMQP settings
    hostname = capo.getstring(
        'edu.nrao.archive.configuration.AmqpServer.hostname')
    username = capo.getstring(
        'edu.nrao.archive.configuration.AmqpServer.username')
    password = capo.getstring(
        'edu.nrao.archive.configuration.AmqpServer.password')

    # connect
    connection = pika.BlockingConnection(
        pika.ConnectionParameters(hostname,
                                  credentials=pika.PlainCredentials(username,
                                                                    password)))
    return connection


#
# Helper class for tracking a workflow:
#
class WorkflowTracker:
    """An encapsulation of using an RPCEvent object to generate an ID
       and receive the working directory and loggingId for a workflow
       launched via wf.

        Usage:
        my_tracker = new WorkflowTracker(current_profile)
        uid_value = my_tracker.prepareWorkflowTracking("runMyWorkflow", <optional email address for completion>)

        # create a wf call as usual, but include another key in the metadata:
        wf_arguments += " -p cliCorrId=" + uid_value
        # send the wf command


        my_tracker.waitForReply()

        wd_name = my_tracker.getWorkingDirectory()
        lid = my_tracker.getLoggingId()

        print("Yadda, yadda, {} yadda {}".format(wd_name,lid))


       """

    def __init__(self, profile):
        self._rpc = RPCEvent(profile=profile, exchange='archive.events', application='commandline')

        self._work_dir_name = None
        self._log_id = None

    def close(self):
        self._rpc.close()

    def prepareWorkflowTracking(self, workflow_name, email=None):

        # Create the event: Versions with and without an email.
        if email is None:
            event = {
                "message": "Tracking Request for" + workflow_name,
                "request": "monitor",
                "logData": {
                    "workflowId": workflow_name
                }
            }
        else:
            event = {
                "message": "Tracking Request for " + workflow_name,
                "request": "monitor",
                "logData": {
                    "workflowId": workflow_name,
                    "email": email
                }
            }

        tracking_key = self._rpc.send(event)
        return tracking_key

    def waitForReply(self):
        # Now, wait for the initial processing and populate our variables
        # This will block until the direct reply arrives.
        answer = self._rpc.get_reply()

        if "working_directory" in answer:
            self._work_dir_name = answer["working_directory"]

        if "logging_id" in answer:
            self._log_id = answer["logging_id"]

    def getWorkingDirectory(self):
        return self._work_dir_name

    def getLoggingId(self):
        return self._log_id

#
# Specific WF usages
#


# CIPL CLI
_RUN_CIPL_DESCRIPTION = """AAT/PPI Calibration Launcher, version {}: This tool initiates the standard CASA Calibration 
Pipeline on the indicated execution blocks.  Note: This tool ignores the ciplRunState Capo property.  It will always
 start the requested calibrations."""


def _make_cipl_parser():
    r"""Command-line CIPL Launch Parser"""
    result = ap.ArgumentParser(description=_RUN_CIPL_DESCRIPTION.format(version),
                           formatter_class=ap.RawTextHelpFormatter)
    result.add_argument('-P', '--profile', action='store', default='',
                    help='profile name to use, e.g. nmtest, mnprod')
    result.add_argument('-C', '--casa', action='store', default='',
                        help='Path to Casa package to use')
    result.add_argument('fileSetIds', type = str, action='store', nargs='+', help='File Set Identifiers to be run through CIPL')
    return result


def launch_calibrations():
    r"""Take given fileSetId(s) and launch the CIPL workflow for each"""

    #parse our arguments, get profile and list of Ids
    args = _make_cipl_parser().parse_args()

    # Shamelessly stolen from epilogue with a twist: allow for explict profile setting via the CL
    if 'CAPO_PROFILE' not in os.environ and '' == args.profile:
        # try to synthesize a profile from our installation root
        profile = os.path.abspath(sys.argv[0]).split(os.path.sep)[-3]
        os.environ['CAPO_PROFILE'] = profile
        print('No CAPO_PROFILE explicitly set, synthesizing {0} by path'.format(str(profile)))
    elif '' != args.profile:
        os.environ['CAPO_PROFILE'] = args.profile

    #
    # Command line we want is:
    #
    # wf -P [test|nmprod] -p workflowName=CiplCalibration -p fileSetIds=arg runCiplCalibrationWorkflow
    #   or
    # wf -P [nmtest|nmprod] -p workflowName=CiplCalibration -p casaHome=arg -p fileSetIds=arg runCiplCalibrationWorkflow
    #
    # The pieces of the string
    clBase = "-p workflowName=CiplCalibration -p processingSite=DSOC -P " #profile here
    clFSIDPiece = " -p fileSetIds=" # fileSetId here
    clTrackPiece = " -p cliCorrId=" # tracking UID here
    if '' != args.casa:
        clCasaPiece = " -p casaHome=" + args.casa
        # insert the casa parameter here, since it is fixed for all fileSetIds
        clFSIDPiece = clCasaPiece + clFSIDPiece

    clFinish = " runCiplCalibrationWorkflow"

    # create a tracking object for these workflows:
    ranger = WorkflowTracker(os.environ['CAPO_PROFILE'])

    #
    # Loop over the provided fileSetIds and send a command for each,
    # waiting for the previous one to report back information before
    # starting the next.
    #
    for id in args.fileSetIds:

        # CIPL itself send emails at start & finish, so it'd be redundant here
        this_uid = ranger.prepareWorkflowTracking('runCiplCalibrationWorkflow')

        wfArgs = clBase + os.environ['CAPO_PROFILE'] + clFSIDPiece + id + clTrackPiece + this_uid + clFinish
        wf(args=wfArgs.split(' '))

        # Wait for information from this workflow before we proceed:
        ranger.waitForReply()

        print("CIPL ({}) for {} running in: {}".format(ranger.getLoggingId(), id, ranger.getWorkingDirectory()))


# CIPL Reprocess CLI
_RUN_CIPL_REPRO_DESCRIPTION = """AAT/PPI Calibration Reprocessing, version {}: This tool initiates a re-run of the CASA
calibration pipeline on the desired directories.  The data are placed in the Spool for the processing and returned to the 
QA area upon successful completion."""


def _make_cipl_reprocess_parser():
    r"""Parser"""
    result = ap.ArgumentParser(description=_RUN_CIPL_REPRO_DESCRIPTION.format(version),
                           formatter_class=ap.RawTextHelpFormatter)
    result.add_argument('-P', '--profile', action='store', default='',
                    help='profile name to use, e.g. test, production')
    result.add_argument('-C', '--casa', action='store', default='',
                        help='Path to Casa package to use')
    result.add_argument('-m', '--main', action='store_true', help='Work in the main spool directory, not the cipl subdirectory')
    result.add_argument('directoriesOfInterest', type = str, action='store', nargs='+', help='Name of the directories to be re-run through CIPL')
    return result


def launch_recalibrations():
    """From a (list of) directory(s) launch the CIPL Reprocessing workflow for each"""

    #parse our arguments, get profile and list of Ids
    args = _make_cipl_reprocess_parser().parse_args()

    # Shamelessly stolen from epilogue with a twist: allow for explict profile setting via the CL
    if 'CAPO_PROFILE' not in os.environ and '' == args.profile:
        # try to synthesize a profile from our installation root
        profile = os.path.abspath(sys.argv[0]).split(os.path.sep)[-3]
        os.environ['CAPO_PROFILE'] = profile
        print('No CAPO_PROFILE explicitly set, synthesizing {0} by path'.format(str(profile)))
    elif '' != args.profile:
        os.environ['CAPO_PROFILE'] = args.profile


    #
    # Command line we want is:
    # (processingSite for message routing)
    #
    # wf -P [nmtest|nmprod] -p workflowName=CiplReprocess -p processingSite=DSOC -p targetDirectory=arg runCiplReprocessingWorkflow
    #
    # The pieces of the string
    clBase = "-p workflowName=CiplReprocess -p processingSite=DSOC -P " #profile here
    clArgPiece = " -p targetDirectory="                    #fileSetId here
    if '' != args.casa:
        clCasaPiece = " -p casaHome=" + args.casa
        # insert the casa parameter here, since it is fixed for all fileSetIds
        clArgPiece = clCasaPiece + clArgPiece
    clArg2Piece = " -p useMainDirectory=true"
    clFinish = " runCiplReprocessingWorkflow"
    #clFinish = " runTestWorkflow"
    #
    # Loop over the provided directories and send a command for each:
    #
    for id in args.directoriesOfInterest:
        #add the extra -p call if we're using the super-directory
        if(args.main):
            totalCommand = clBase+ os.environ['CAPO_PROFILE'] +clArgPiece+ id +clArg2Piece+clFinish
        else:
            totalCommand = clBase+ os.environ['CAPO_PROFILE'] +clArgPiece+ id +clFinish
        wf(args=totalCommand.split(' '))
        time.sleep(0.25) # 1/4 second deplay between workflow launches for saftey


# SDM reingestion CLI
_RUN_SDM_REINGEST_DESCRIPTION = """AAT/PPI SDM Re-ingestion, version {}: This tool initiates a download of the requested 
Science Data Models, following which the files are re-parsed for an update of the Archive metadata database."""


def _make_sdm_reingest_parser():
    result = ap.ArgumentParser(description=_RUN_SDM_REINGEST_DESCRIPTION.format(version),
                               formatter_class=ap.RawTextHelpFormatter)
    result.add_argument('-P', '--profile', action='store',
                        help='profile name to use, e.g. test, production')
    result.add_argument('fileSetId', type = str, action='store', nargs='+',
                        help='Name of the fileset whose SDM is to be reingested')
    return result


def launch_reingest():
    """Given a fileset ID, pull the (presumably altered by QA) SDM and re-run ingestion """

    # TODO once this is working:
    # - accept a list of fileset IDs
    # - accept a telescope name

    args = _make_sdm_reingest_parser().parse_args()

    # command line:
    # Test cases:
    # 1. no SDM source directory supplied
    #        -- should ultimately fail if data not present under archive-ingestion.SDMPath
    #    wf -P [test|nmprod] -p workflowName=Reingestion -p telescope=EVLA \
    #       -p dataSetId=<FSID> runSdmReingestWorkflow
    #    example: wf -P test -p workflowName=Reingestion -p telescope=EVLA \
    #       -p dataSetId=3_4.sb31082669.eb31333837.57311.64081266204 \
    #       -p fileSetIds=[3_4.sb31082669.eb31333837.57311.64081266204] \
    #           runSdmReingestWorkflow
    # 2. user provides SDM source directory
    #    wf -P [test|nmprod] -p workflowName=Reingestion -p telescope=EVLA -p dataSetId=<FSID> \
    #       -p SDMPath=</path/to/SDM> runSdmIngestionWorkflow
    #    example:  wf -P test -p workflowName=Reingestion -p telescope=EVLA \
    #       -p dataSetId=3_4.sb31082669.eb31333837.57311.64081266204 \
    #       -p fileSetIds=[3_4.sb31082669.eb31333837.57311.64081266204] \
    #       -p SDMPath=/lustre/aoc/cluster/pipeline/test/spool/3_4_2015_10_16_T21_58_09.207/rawdata \
    #       runSdmReingestWorkflow

    reArg0 = "-p dataSetId="
    reArg1 = "-p fileSetIds=["
    reArg2 = "] -p workflowName=Reingestion "
    reArg3 = "-p telescope=EVLA"
    reLast = " runSdmReingestWorkflow"
    command_line = "-P " + args.profile + reArg0 + args.fileSetId + reArg2
    if len(args) < 5:
        # the usual ingestion case: grab SDM from standard archive location
        command_line += reArg3
    else:
        # manual or auto re-ingest: grab SDM from user-supplied location0
        # (use case: SDM was corrupt and fixed manually, now needs to be reingested)
        reArg4 = " -p SDMPath=" + args.sdmSubdir
        command_line = "-P " + args.profile + reArg0  + args.fileSetId + reArg1 + args.fileSetId + reArg2 + reArg3 + reArg4
        command_line += reLast

    LOG = logging.getLogger(__name__)
    handler = LogHandler(profile='test', application='test-app',
                         level=logging.DEBUG)
    LOG.addHandler(handler)
    #OG.debug("Using this command line: {}" %command_line)
    # in case that syntax is wrong:
    LOG.debug("Using command line '%s'" %command_line)

    wf(args=command_line.split(' '))


# RestoreToCache Workflow CLI
_RUN_RESTORE_DESCRIPTION = """AAT/PPI Restore to Cache, version {}:  This tool takes a FileSetId, and the name of a 
calibration products tar file.  It will then initiate a CASA Restore Workflow.  The resultant Calibrated Measurement Set 
is placed into the Cache as defined by the Capo settings provided."""


def _make_r2c_parser():
    r"""Command-line CIPL Launch Parser"""
    result = ap.ArgumentParser(description=_RUN_RESTORE_DESCRIPTION.format(version),
                               formatter_class=ap.RawTextHelpFormatter)
    result.add_argument('-P', '--profile', action='store', default='',
                        help='profile name to use, e.g. nmtest, mnprod')
    result.add_argument('-C', '--casa', action='store', default='',
                        help='Path to Casa package to use')
    result.add_argument('-E', '--email', action='store', default=None,
                        help='Optional email address for a completion message')
    # Placeholder, workflow can't do this yet.
    # result.add_argument('-D', '--destination', action='store', default='', help='Override location for CMS delivery')
    result.add_argument('fileSetId', type = str, action='store', help='File Set Identifier to be Restored')
    result.add_argument('calFile', type=str, action='store', help='Calibration File to use in the Restore')
    return result


def launch_restore_to_cache():
    r"""Take given fileSetId(s) and launch the CIPL workflow for each"""

    # parse our arguments, get profile and list of Ids
    args = _make_r2c_parser().parse_args()

    # Shamelessly stolen from epilogue with a twist: allow for explict profile setting via the CL
    if 'CAPO_PROFILE' not in os.environ and '' == args.profile:
        # try to synthesize a profile from our installation root
        profile = os.path.abspath(sys.argv[0]).split(os.path.sep)[-3]
        os.environ['CAPO_PROFILE'] = profile
        print('No CAPO_PROFILE explicitly set, synthesizing {0} by path'.format(str(profile)))
    elif '' != args.profile:
        os.environ['CAPO_PROFILE'] = args.profile

    # Obtain the Project Code from the fileSetId value (careful of VLASSA.B format)
    fsid = args.fileSetId
    if fsid.startswith("VLASS"):
        components = fsid.split('.')
        projectCode = components[0] + '.' + components[1]  # Grab the first two substrings
    else:
        projectCode = fsid.split('.')[0]  # Only up to the first .

    # Send an event to initiate workflow monitoring:
    my_tracker = WorkflowTracker(args.profile)
    monitor_key = my_tracker.prepareWorkflowTracking("runRestoreWorkflow", email=args.email)

    # Launch the workflow
    launch_restore(args, projectCode, monitor_key)

    # Now wait for information:
    my_tracker.waitForReply()

    print("Restore ({}) running in: {}".format(my_tracker.getLoggingId(), my_tracker.getWorkingDirectory()))


def launch_restore(args, project_code, tracking_key):
    #
    # An Example Command Line from workflow testing:
    #
    # wf -p processingSite=DSOC -p workflowName=RestoreWorkflow -p telescope=EVLA
    #   -p calFile=16A-439_2016_10_09_T14_48_37.675.tar
    #   -p fileSetIds=16A-439.sb32267588.eb32942323.57670.592283148144
    #   -p projectCodeOrDataType=16A-439
    #   -p deliveryFormat=CMS runRestoreWorkflow
    #
    # Build the wf command line, starting with all the fixed pieces:
    arguments = "-P " + os.environ['CAPO_PROFILE']
    # get the exchange from our requested profile (since we might be sending this to the non-default manager)
    configuration = CapoConfig(profile=os.environ['CAPO_PROFILE'])
    arguments += " -e " +  configuration.get('edu.nrao.archive.workflow.config.WorkflowManagerSettings.workflowCommandsExchangeName')
    # Required Parameters
    arguments += " -p workflowName=RestoreWorkflow -p processingSite=DSOC -p deliveryFormat=CMS -p telescope=EVLA"
    arguments += " -p fileSetIds=" + args.fileSetId
    arguments += " -p calFile=" + args.calFile
    arguments += " -p projectCodeOrDataType=" + project_code
    arguments += " -p cliCorrId=" + tracking_key
    # The Optional Parameters:
    if '' != args.casa:
        arguments += " -p casaHome=" + args.casa
    # Option not available yet, so don't display or use it
    # if '' != args.destination:
    #     arguments += " -p deliveryDirectory=" + args.destination
    # The Finale
    arguments += " runRestoreWorkflow"
    # Build our command line and execute:
    wf(args=arguments.split(' '))
    # Talk to the user, these CLIs are bad on feedback
    print("Restoring a CMS of {} to the cache using calibration {}.\n".format(args.fileSetId, args.calFile))


